# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import sphinx_book_theme
import sphinx_book_theme
import os
import sys
sys.path.insert(0, os.path.abspath('..'))
print(f"Root dir: {sys.path[0]}")

# -- Project information -----------------------------------------------------

project = 'redbird'
copyright = '2022, Mikael Koli'
author = 'Mikael Koli'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc', 
    'sphinx.ext.coverage', 
    'sphinx.ext.napoleon',
    'sphinx_book_theme',
    'sphinx_copybutton',
]
rst_prolog = """
.. include:: <s5defs.txt>

"""



# Extension settings

napoleon_google_docstring = True
napoleon_numpy_docstring = True

autodoc_typehints = 'none'
autodoc_member_order = 'bysource'

# Add any paths that contain templates here, relative to this directory.
#templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

html_logo = "logo.png"
html_theme_options = {
    "repository_url": "https://github.com/Miksus/red-bird",
    "use_repository_button": True,
    "repository_branch": "master",
    "use_issues_button": True,
    "use_download_button": True,
    "use_fullscreen_button": True,
    #"use_edit_page_button": True,
}
#html_sidebars = {}


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_title = "Repository patterns for Python"
html_theme = 'sphinx_book_theme'
html_favicon = 'favicon.ico'


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

html_css_files = [
    'css/types.css',
    'css/colors.css',
    "css/custom.css",
]


import pytest
from redbird.sql import Table

def test_multi(table_multiindex, table_singleindex):
    # Single index
    assert table_singleindex[["a", "b"]] == [
        {"index_1": "a", "column_1": "Jack", "column_2": 10},
        {"index_1": "b", "column_1": "James", "column_2": 8},
    ]

    # Multi indices
    assert table_multiindex["a"] == [
        {"index_1": "a", "index_2": 1, "column_1": "Jack", "column_2": 10},
        {"index_1": "a", "index_2": 2, "column_1": "John", "column_2": 9},
    ]

    assert table_multiindex[[("a", 1), ("b", 2)]] == [
        {"index_1": "a", "index_2": 1, "column_1": "Jack", "column_2": 10},
        {"index_1": "b", "index_2": 2, "column_1": "Johnny", "column_2": 7},
    ]
    assert table_multiindex[["a", "b"]] == [
        {"index_1": "a", "index_2": 1, "column_1": "Jack", "column_2": 10},
        {"index_1": "a", "index_2": 2, "column_1": "John", "column_2": 9},
        {"index_1": "b", "index_2": 1, "column_1": "James", "column_2": 8},
        {"index_1": "b", "index_2": 2, "column_1": "Johnny", "column_2": 7},
    ]

def test_single_item(table_multiindex, table_singleindex):
    # Single index
    assert table_singleindex["a"] == {"index_1": "a", "column_1": "Jack", "column_2": 10}

    # Multiple indices
    assert table_multiindex[("a", 1)] == {"index_1": "a", "index_2": 1, "column_1": "Jack", "column_2": 10}
    assert table_multiindex["a", 1] == {"index_1": "a", "index_2": 1, "column_1": "Jack", "column_2": 10}

def test_slice(table_multiindex, table_singleindex):
    # Single index
    assert table_singleindex["a":"b"] == [
        {"index_1": "a", "column_1": "Jack", "column_2": 10},
        {"index_1": "b", "column_1": "James", "column_2": 8},
    ]
    assert table_singleindex[:"b"] == [
        {"index_1": "a", "column_1": "Jack", "column_2": 10},
        {"index_1": "b", "column_1": "James", "column_2": 8},
    ]


    # Multi-indices
    assert table_multiindex[:] == [
        {"index_1": "a", "index_2": 1, "column_1": "Jack", "column_2": 10},
        {"index_1": "a", "index_2": 2, "column_1": "John", "column_2": 9},
        {"index_1": "b", "index_2": 1, "column_1": "James", "column_2": 8},
        {"index_1": "b", "index_2": 2, "column_1": "Johnny", "column_2": 7},
        {"index_1": "c", "index_2": 1, "column_1": "Jimmy", "column_2": 8},
        {"index_1": "c", "index_2": 2, "column_1": "Jim", "column_2": 7},
        {"index_1": "d", "index_2": 1, "column_1": "Bon", "column_2": 99},
        {"index_1": "d", "index_2": 2, "column_1": "Bam", "column_2": 98},
    ]

    assert table_multiindex["a":"c"] == [
        {"index_1": "a", "index_2": 1, "column_1": "Jack", "column_2": 10},
        {"index_1": "a", "index_2": 2, "column_1": "John", "column_2": 9},
        {"index_1": "b", "index_2": 1, "column_1": "James", "column_2": 8},
        {"index_1": "b", "index_2": 2, "column_1": "Johnny", "column_2": 7},
        {"index_1": "c", "index_2": 1, "column_1": "Jimmy", "column_2": 8},
        {"index_1": "c", "index_2": 2, "column_1": "Jim", "column_2": 7},
    ]

    assert table_multiindex["b":] == [
        {"index_1": "b", "index_2": 1, "column_1": "James", "column_2": 8},
        {"index_1": "b", "index_2": 2, "column_1": "Johnny", "column_2": 7},
        {"index_1": "c", "index_2": 1, "column_1": "Jimmy", "column_2": 8},
        {"index_1": "c", "index_2": 2, "column_1": "Jim", "column_2": 7},
        {"index_1": "d", "index_2": 1, "column_1": "Bon", "column_2": 99},
        {"index_1": "d", "index_2": 2, "column_1": "Bam", "column_2": 98},
    ]
    assert table_multiindex[:"b"] == [
        {"index_1": "a", "index_2": 1, "column_1": "Jack", "column_2": 10},
        {"index_1": "a", "index_2": 2, "column_1": "John", "column_2": 9},
        {"index_1": "b", "index_2": 1, "column_1": "James", "column_2": 8},
        {"index_1": "b", "index_2": 2, "column_1": "Johnny", "column_2": 7},
    ]
    assert table_multiindex[:"0"] == []
    assert table_multiindex["x":] == []
    assert table_multiindex["x":"z"] == []

def test_getitem_error(table_multiindex):

    with pytest.raises(KeyError):
        # Not found, tuple (second key)
        table_multiindex[("a", "not_found")]

    with pytest.raises(KeyError):
        # Not found, range
        table_multiindex["not_found"]

    with pytest.raises(KeyError):
        # Not found, tuple (first key)
        table_multiindex[("not_found", 2)]

    with pytest.raises(KeyError):
        # Not found, list of tuples
        table_multiindex[[("not_found", 2)]]

    with pytest.raises(IndexError):
        # Too long for index
        table_multiindex[("a", 2, "long")]

    with pytest.raises(IndexError):
        # Varying sizes
        table_multiindex[[("a", 2), ("a",)]]

    with pytest.raises(IndexError):
        # Varying sizes
        table_multiindex[[("a", 2), "a"]]

    with pytest.raises(TypeError):
        # Tuple in tuple
        table_multiindex[("a", 2), "a"]

    with pytest.raises(ValueError):
        # Tuple in tuple
        table_multiindex[::2]

def test_getitem_no_primary_key(engine):
    tbl = Table(name="getitems_singleindex", bind=engine)
    tbl.create([
        {"name": "index_1", "type_": str, "primary_key": False},
        {"name": "column_1", "type_": str, "primary_key": False},
        {"name": "column_2", "type_": int, "primary_key": False},
    ])
    tbl.insert([
        {"index_1": "a", "column_1": "Jack", "column_2": 10},
        {"index_1": "b", "column_1": "James", "column_2": 8},
        {"index_1": "c", "column_1": "Jimmy", "column_2": 8},
    ])
    with pytest.raises(TypeError):
        tbl[("a", 1)]




import importlib

def get_import_error(name:str):
    raise ModuleNotFoundError(f"No module named '{name}'")

class _Missing_Package:
    def __init__(self, name:str) -> None:
        self.__name__ = name

    def __setattr__(self, name, value):
        if name == "__name__":
            super().__setattr__(name, value)
            return
        raise get_import_error(self.__name__)

    def __getattr__(self, name):
        raise get_import_error(self.__name__)
    

def import_optional(name:str):
    try:
        return importlib.import_module(name)
    except ModuleNotFoundError:
        return _Missing_Package(name)

def import_exists(name:str):
    return not isinstance(import_optional(name), _Missing_Package)


import pytest
from datetime import date

from redbird.sql import insert, execute, select

def test_insert_dict(engine):
    sqlalchemy = pytest.importorskip("sqlalchemy")
    insert(
        {'id': 'a', 'name': 'Johnny', 'birth_date': date(2000, 1, 1), 'score': 100}, 
        table="empty", 
        bind=engine
    )
    assert [
        {'id': 'a', 'name': 'Johnny', 'birth_date': '2000-01-01', 'score': 100}, 
    ] == list(select("select * from empty", bind=engine))

def test_insert_list_dicts(engine):
    sqlalchemy = pytest.importorskip("sqlalchemy")
    insert(
        [
            {'id': 'a', 'name': 'Johnny', 'birth_date': date(2000, 1, 1), 'score': 100},
            {'id': 'b', 'name': 'James', 'birth_date': date(2020, 1, 1), 'score': 200},
        ], 
        table="empty", 
        bind=engine
    )
    assert [
        {'id': 'a', 'name': 'Johnny', 'birth_date': '2000-01-01', 'score': 100}, 
        {'id': 'b', 'name': 'James', 'birth_date': '2020-01-01', 'score': 200},
    ] == list(select("select * from empty", bind=engine))


import pytest
from datetime import date, datetime

from redbird.sql import insert, select, delete, update, count, execute, Table
from redbird.oper import in_, between, equal

def test_select_all(engine):
    results = select(table="populated", bind=engine)
    assert [
        {'id': 'a', 'name': 'Jack', 'birth_date': date(2000, 1, 1), 'score': 100},
        {'id': 'b', 'name': 'John', 'birth_date': date(1990, 1, 1), 'score': 200},
        {'id': 'c', 'name': 'James', 'birth_date': date(2020, 1, 1), 'score': 300},
    ] == list(results)

@pytest.mark.parametrize("how", [
    "raw string",
    "expression"
])
def test_select_parametize(engine, how):
    sqlalchemy = pytest.importorskip("sqlalchemy")
    if how == "raw string":
        qry = "select * from populated where name = :name"
    elif how == "expression":
        qry = sqlalchemy.text("select * from populated where name = :name")
        
    results = select(qry, bind=engine, parameters={"name": "Jack"})
    assert [
        {'id': 'a', 'name': 'Jack', 'birth_date': '2000-01-01', 'score': 100},
        #{'id': 'b', 'name': 'John', 'birth_date': '1990-01-01', 'score': 200},
        #{'id': 'c', 'name': 'James', 'birth_date': '2020-01-01', 'score': 300},
    ] == list(results)

@pytest.mark.parametrize("how", [
    "raw string",
    'native',
    "expression",
])
def test_select_equal(engine, how):
    import sqlalchemy

    if how == "raw string":
        qry = "SELECT * FROM populated where name == 'John'"
    elif how == "native":
        qry = {"name": "John"}
    elif how == "expression":
        qry = sqlalchemy.Column("name") == "John"
    results = select(qry, bind=engine, table="populated")
    assert [
        #{'id': 'a', 'name': 'Jack', 'birth_date': date(2000, 1, 1), 'score': 100},
        {'id': 'b', 'name': 'John', 'birth_date': date(1990, 1, 1), 'score': 200},
        #{'id': 'c', 'name': 'James', 'birth_date': date(2020, 1, 1), 'score': 300},
    ] == list(results)

@pytest.mark.parametrize("how", [
    "raw string",
    "expression",
    "operation",
])
def test_select_in(engine, how):
    import sqlalchemy

    if how == "raw string":
        qry = "SELECT * FROM populated where name in ('Jack', 'John')"
    elif how == "expression":
        qry = sqlalchemy.Column("name").in_(["Jack", "John"])
    elif how == "operation":
        qry = {"name": in_(["Jack", "John"])}
    results = select(qry, bind=engine, table="populated")
    assert [
        {'id': 'a', 'name': 'Jack', 'birth_date': date(2000, 1, 1), 'score': 100},
        {'id': 'b', 'name': 'John', 'birth_date': date(1990, 1, 1), 'score': 200},
        #{'id': 'c', 'name': 'James', 'birth_date': date(2020, 1, 1), 'score': 300},
    ] == list(results)

@pytest.mark.parametrize("how", [
    "raw string",
    'native',
    "expression",
    "operation",
])
def test_select_range(engine, how):
    import sqlalchemy
    start = 100
    end = 220
    if how == "raw string":
        qry = f"SELECT * FROM populated where score between {start} and {end}"
    elif how == "native":
        qry = {"score": slice(start, end)}
    elif how == "expression":
        qry = sqlalchemy.Column("score").between(start, end)
    elif how == "operation":
        qry = {"score": between(start, end)}
    results = select(qry, bind=engine, table="populated")
    assert [
        {'id': 'a', 'name': 'Jack', 'birth_date': date(2000, 1, 1), 'score': 100},
        {'id': 'b', 'name': 'John', 'birth_date': date(1990, 1, 1), 'score': 200},
        #{'id': 'c', 'name': 'James', 'birth_date': date(2020, 1, 1), 'score': 300},
    ] == list(results)

@pytest.mark.parametrize("how", [
    "raw string", "statement"
])
def test_select_without_table(engine, how):
    import sqlalchemy
    if how == "raw string":
        qry = "SELECT * FROM populated where name in ('Jack', 'John')"
    elif how == "statement":
        tbl = Table("populated", bind=engine).object
        qry = sqlalchemy.select(tbl).where(tbl.columns['name'].in_(["Jack", "John"]))

    results = select(qry, bind=engine)
    
    if how == "statement":
        assert [
            {'id': 'a', 'name': 'Jack', 'birth_date': date(2000, 1, 1), 'score': 100},
            {'id': 'b', 'name': 'John', 'birth_date': date(1990, 1, 1), 'score': 200},
            #{'id': 'c', 'name': 'James', 'birth_date': '2020-01-01', 'score': 300},
        ] == list(results)
    else:
        assert [
            {'id': 'a', 'name': 'Jack', 'birth_date': '2000-01-01', 'score': 100},
            {'id': 'b', 'name': 'John', 'birth_date': '1990-01-01', 'score': 200},
            #{'id': 'c', 'name': 'James', 'birth_date': '2020-01-01', 'score': 300},
        ] == list(results)

def test_select_dict(engine):
    results = select({"name": "James", "birth_date": date(2020, 1, 1)}, table="populated", bind=engine)
    assert [
        #{'id': 'a', 'name': 'Jack', 'birth_date': '2000-01-01', 'score': 100},
        #{'id': 'b', 'name': 'John', 'birth_date': '1990-01-01', 'score': 200},
        {'id': 'c', 'name': 'James', 'birth_date': date(2020, 1, 1), 'score': 300},
    ] == list(results)

def test_select_expr(engine):
    sqlalchemy = pytest.importorskip("sqlalchemy")
    results = select((sqlalchemy.Column("name") == "James") & (sqlalchemy.Column("birth_date") == date(2020, 1, 1)), table="populated", bind=engine)
    assert [
        #{'id': 'a', 'name': 'Jack', 'birth_date': '2000-01-01', 'score': 100},
        #{'id': 'b', 'name': 'John', 'birth_date': '1990-01-01', 'score': 200},
        {'id': 'c', 'name': 'James', 'birth_date': date(2020, 1, 1), 'score': 300},
    ] == list(results)


import pytest
from redbird.utils.importing import import_optional

def test_import():
    pkg: pytest = import_optional("pytest")
    assert pkg is pytest
    assert pkg.main is pytest.main

def test_import_fail():
    pkg = import_optional("redbird_missing_package")
    assert pkg.__name__ == "redbird_missing_package"
    with pytest.raises(ModuleNotFoundError):
        pkg.session
    with pytest.raises(ModuleNotFoundError):
        pkg.session = "value"



import pytest
from datetime import date, datetime

from redbird.sql import insert, select, delete, update, count, execute, Table
from redbird.oper import in_, between

def test_update_all(engine):
    sqlalchemy = pytest.importorskip("sqlalchemy")

    assert update(
        {},
        {'score': -999},
        table="populated", 
        bind=engine
    ) == 3

    assert [
        {'id': 'a', 'name': 'Jack', 'birth_date': '2000-01-01', 'score': -999},
        {'id': 'b', 'name': 'John', 'birth_date': '1990-01-01', 'score': -999},
        {'id': 'c', 'name': 'James', 'birth_date': '2020-01-01', 'score': -999},
    ] == list(select("select * from populated", bind=engine))

@pytest.mark.parametrize("how", [
    "dict", "expressions"
])
def test_update(engine, how):
    sqlalchemy = pytest.importorskip("sqlalchemy")

    if how == "dict":
        where = {'name': 'Jack', 'birth_date': date(2000, 1, 1)}
    elif how == "expressions":
        where = (sqlalchemy.Column("name") == "Jack") & (sqlalchemy.Column("birth_date") == date(2000, 1, 1))

    assert update(
        where,
        {'score': 0},
        table="populated", 
        bind=engine
    ) == 1
    assert [
        {'id': 'a', 'name': 'Jack', 'birth_date': '2000-01-01', 'score': 0},
        {'id': 'b', 'name': 'John', 'birth_date': '1990-01-01', 'score': 200},
        {'id': 'c', 'name': 'James', 'birth_date': '2020-01-01', 'score': 300},
    ] == list(select("select * from populated", bind=engine))


import pytest
from redbird.sql import Table


@pytest.fixture()
def table_singleindex(engine) -> Table:
    tbl = Table(name="getitems_singleindex", bind=engine)
    tbl.create([
        {"name": "index_1", "type_": str, "primary_key": True},
        {"name": "column_1", "type_": str, "primary_key": False},
        {"name": "column_2", "type_": int, "primary_key": False},
    ])
    tbl.insert([
        {"index_1": "a", "column_1": "Jack", "column_2": 10},
        {"index_1": "b", "column_1": "James", "column_2": 8},
        {"index_1": "c", "column_1": "Jimmy", "column_2": 8},
    ])
    return tbl

@pytest.fixture()
def table_multiindex(engine) -> Table:
    tbl = Table(name="getitems", bind=engine)
    tbl.create([
        {"name": "index_1", "type_": str, "primary_key": True},
        {"name": "index_2", "type_": int, "primary_key": True},
        {"name": "column_1", "type_": str, "primary_key": False},
        {"name": "column_2", "type_": int, "primary_key": False},
    ])
    tbl.insert([
        {"index_1": "a", "index_2": 1, "column_1": "Jack", "column_2": 10},
        {"index_1": "a", "index_2": 2, "column_1": "John", "column_2": 9},
        {"index_1": "b", "index_2": 1, "column_1": "James", "column_2": 8},
        {"index_1": "b", "index_2": 2, "column_1": "Johnny", "column_2": 7},
        {"index_1": "c", "index_2": 1, "column_1": "Jimmy", "column_2": 8},
        {"index_1": "c", "index_2": 2, "column_1": "Jim", "column_2": 7},
        {"index_1": "d", "index_2": 1, "column_1": "Bon", "column_2": 99},
        {"index_1": "d", "index_2": 2, "column_1": "Bam", "column_2": 98},
    ])
    return tbl


from fastapi import FastAPI
from pydantic import BaseModel
from redbird.repos import MemoryRepo

app = FastAPI()

class Item(BaseModel):
    id: int
    name: str

repo = MemoryRepo(model=Item, id_field="id")

@app.post("/items", description="Create an item")
def create_item(item: Item):
    repo.add(item)

@app.get("/items", description="Get all items")
def get_items():
    return repo.filter_by().all()

@app.get("/items/{item_id}", description="Get item by ID")
def get_item(item_id: int):
    return repo[item_id]

@app.patch("/items/{item_id}", description="Update an item")
def update_item(item_id: int, values: dict):
    repo[item_id] = values

@app.delete("/items/{item_id}", description="Delete an item")
def delete_item(item_id: int):
    del repo[item_id]



import pytest
import redbird

def test_build(request):
    expect_build = request.config.getoption('is_build')
    if not expect_build:
        assert redbird.version == '0.0.0.UNKNOWN'
    else:
        assert redbird.version != '0.0.0.UNKNOWN'


from typing import Optional
from pydantic import BaseModel
from datetime import date
import pytest

from redbird.sql import Table, create_table, execute, select

@pytest.mark.parametrize("column_type", [
    "string",
    "sqlalchemy",
    "dict",
    "list-dicts"
])
def test_create(engine, column_type):
    import sqlalchemy
    if column_type == "string":
        columns = ["col1", "col2"]
    elif column_type == "dict":
        columns = {"col1": str, "col2": sqlalchemy.String()}
    elif column_type == "sqlalchemy":
        columns = [
            sqlalchemy.Column("col1", type_=sqlalchemy.String()),
            sqlalchemy.Column("col2", type_=sqlalchemy.String())
        ]
    elif column_type == "list-dicts":
        columns = [
            {"name": "col1", "type_": str},
            {"name": "col2", "type_": sqlalchemy.String()},
        ]
    create_table(table="mytable", columns=columns, bind=engine)

    res = select("PRAGMA table_info(mytable);", bind=engine)
    assert [
        {'cid': 0, 'name': 'col1', 'type': 'VARCHAR', 'notnull': 0, 'dflt_value': None, 'pk': 0},
        {'cid': 1, 'name': 'col2', 'type': 'VARCHAR', 'notnull': 0, 'dflt_value': None, 'pk': 0},
    ] == list(res)

def test_drop(engine):
    tbl = Table("mytable", bind=engine)

    tbl.create(["col1", "col2"])
    res = select("PRAGMA table_info(mytable);", bind=engine)
    assert [
        {'cid': 0, 'name': 'col1', 'type': 'VARCHAR', 'notnull': 0, 'dflt_value': None, 'pk': 0},
        {'cid': 1, 'name': 'col2', 'type': 'VARCHAR', 'notnull': 0, 'dflt_value': None, 'pk': 0},
    ] == list(res)

    tbl.drop()
    res = select("PRAGMA table_info(mytable);", bind=engine)
    assert [] == list(res)

def test_exists(engine):
    tbl = Table("mytable", bind=engine)
    assert not tbl.exists()
    tbl.create(["col1", "col2"])
    assert tbl.exists()
    tbl.drop()
    assert not tbl.exists()

def test_create_already_exists(engine):
    import sqlalchemy
    create_table(table="mytable", columns=["col1", "col2"], bind=engine)

    res = select("PRAGMA table_info(mytable);", bind=engine)
    assert [
        {'cid': 0, 'name': 'col1', 'type': 'VARCHAR', 'notnull': 0, 'dflt_value': None, 'pk': 0},
        {'cid': 1, 'name': 'col2', 'type': 'VARCHAR', 'notnull': 0, 'dflt_value': None, 'pk': 0},
    ] == list(res)

    with pytest.raises(sqlalchemy.exc.OperationalError):
        create_table(table="mytable", columns=["col1", "col2", "col3"], bind=engine)
    
    create_table(table="mytable", columns=["col1", "col2", "col3"], exist_ok=True, bind=engine)

def test_create_model(engine):
    class MyModel(BaseModel):
        name: str
        score: int = 999
        birth_date: date
        color: Optional[str]

    tbl = Table("mytable", bind=engine)
    tbl.create_from_model(MyModel)
    res = select("PRAGMA table_info(mytable);", bind=engine)
    assert [
        {'cid': 0, 'name': 'name', 'type': 'VARCHAR', 'notnull': 1, 'dflt_value': None, 'pk': 0},
        {'cid': 1, 'name': 'score', 'type': 'INTEGER', 'notnull': 0, 'dflt_value': None, 'pk': 0},
        {'cid': 2, 'name': 'birth_date', 'type': 'DATE', 'notnull': 1, 'dflt_value': None, 'pk': 0},
        {'cid': 3, 'name': 'color', 'type': 'VARCHAR', 'notnull': 0, 'dflt_value': None, 'pk': 0},
    ] == list(res)

    tbl.insert({"name": "Jack", "birth_date": date(2000, 1, 1)})
    assert list(tbl.select()) == [{'name': 'Jack', 'score': 999, 'birth_date': date(2000, 1, 1), 'color': None}]

def test_create_model_primary_key(engine):
    class MyModel(BaseModel):
        myid: str

    tbl = Table("mytable", bind=engine)
    tbl.create_from_model(MyModel, primary_column="myid")
    res = select("PRAGMA table_info(mytable);", bind=engine)
    assert [
        {'cid': 0, 'name': 'myid', 'type': 'VARCHAR', 'notnull': 1, 'dflt_value': None, 'pk': 1},
    ] == list(res)

def test_create_model_multi_primary_key(engine):
    class MyModel(BaseModel):
        myid1: str
        myid2: int

    tbl = Table("mytable", bind=engine)
    tbl.create_from_model(MyModel, primary_column=("myid1", "myid2"))
    res = select("PRAGMA table_info(mytable);", bind=engine)
    assert [
        {'cid': 0, 'name': 'myid1', 'type': 'VARCHAR', 'notnull': 1, 'dflt_value': None, 'pk': 1},
        {'cid': 1, 'name': 'myid2', 'type': 'INTEGER', 'notnull': 1, 'dflt_value': None, 'pk': 2},
    ] == list(res)

def test_create_dict(engine):

    tbl = Table("mytable", bind=engine)
    tbl.create({
        "mycol_1": str, 
        "mycol_2": int, 
        "mycol_3": date,
    })
    res = select("PRAGMA table_info(mytable);", bind=engine)
    assert [
        {'cid': 0, 'name': 'mycol_1', 'type': 'VARCHAR', 'notnull': 0, 'dflt_value': None, 'pk': 0},
        {'cid': 1, 'name': 'mycol_2', 'type': 'INTEGER', 'notnull': 0, 'dflt_value': None, 'pk': 0},
        {'cid': 2, 'name': 'mycol_3', 'type': 'DATE', 'notnull': 0, 'dflt_value': None, 'pk': 0},
    ] == list(res)

def test_create_list(engine):
    sqlalchemy = pytest.importorskip("sqlalchemy")
    tbl = Table("mytable", bind=engine)
    tbl.create([
        sqlalchemy.Column("mycol_1", sqlalchemy.VARCHAR, primary_key=True),
        sqlalchemy.Column("mycol_2", sqlalchemy.INTEGER, nullable=False),
        sqlalchemy.Column("mycol_3", sqlalchemy.DATE),
    ])
    res = select("PRAGMA table_info(mytable);", bind=engine)
    assert [
        {'cid': 0, 'name': 'mycol_1', 'type': 'VARCHAR', 'notnull': 1, 'dflt_value': None, 'pk': 1},
        {'cid': 1, 'name': 'mycol_2', 'type': 'INTEGER', 'notnull': 1, 'dflt_value': None, 'pk': 0},
        {'cid': 2, 'name': 'mycol_3', 'type': 'DATE', 'notnull': 0, 'dflt_value': None, 'pk': 0},
    ] == list(res)



import pytest

@pytest.mark.parametrize("qry,results", [
    pytest.param(
        "b", [
            {"index_1": "a", "index_2": 1, "column_1": "Jack", "column_2": 10},
            {"index_1": "a", "index_2": 2, "column_1": "John", "column_2": 9},
            #{"index_1": "b", "index_2": 1, "column_1": "James", "column_2": 8},
            #{"index_1": "b", "index_2": 2, "column_1": "Johnny", "column_2": 7},
            {"index_1": "c", "index_2": 1, "column_1": "Jimmy", "column_2": 8},
            {"index_1": "c", "index_2": 2, "column_1": "Jim", "column_2": 7},
            {"index_1": "d", "index_2": 1, "column_1": "Bon", "column_2": 99},
            {"index_1": "d", "index_2": 2, "column_1": "Bam", "column_2": 98},
        ],
        id="range item (string)"
    ),
    pytest.param(
        ["b", "d"], [
            {"index_1": "a", "index_2": 1, "column_1": "Jack", "column_2": 10},
            {"index_1": "a", "index_2": 2, "column_1": "John", "column_2": 9},
            #{"index_1": "b", "index_2": 1, "column_1": "James", "column_2": 8},
            #{"index_1": "b", "index_2": 2, "column_1": "Johnny", "column_2": 7},
            {"index_1": "c", "index_2": 1, "column_1": "Jimmy", "column_2": 8},
            {"index_1": "c", "index_2": 2, "column_1": "Jim", "column_2": 7},
            #{"index_1": "d", "index_2": 1, "column_1": "Bon", "column_2": 99},
            #{"index_1": "d", "index_2": 2, "column_1": "Bam", "column_2": 98},
        ],
        id="range item (list of strings)"
    ),
    pytest.param(
        ("b",), 
        [
            {"index_1": "a", "index_2": 1, "column_1": "Jack", "column_2": 10},
            {"index_1": "a", "index_2": 2, "column_1": "John", "column_2": 9},
            #{"index_1": "b", "index_2": 1, "column_1": "James", "column_2": 8},
            #{"index_1": "b", "index_2": 2, "column_1": "Johnny", "column_2": 7},
            {"index_1": "c", "index_2": 1, "column_1": "Jimmy", "column_2": 8},
            {"index_1": "c", "index_2": 2, "column_1": "Jim", "column_2": 7},
            {"index_1": "d", "index_2": 1, "column_1": "Bon", "column_2": 99},
            {"index_1": "d", "index_2": 2, "column_1": "Bam", "column_2": 98},
        ],
        id="range item (tuple)"
    ),
    pytest.param(
        [("b",), ("d",)], 
        [
            {"index_1": "a", "index_2": 1, "column_1": "Jack", "column_2": 10},
            {"index_1": "a", "index_2": 2, "column_1": "John", "column_2": 9},
            #{"index_1": "b", "index_2": 1, "column_1": "James", "column_2": 8},
            #{"index_1": "b", "index_2": 2, "column_1": "Johnny", "column_2": 7},
            {"index_1": "c", "index_2": 1, "column_1": "Jimmy", "column_2": 8},
            {"index_1": "c", "index_2": 2, "column_1": "Jim", "column_2": 7},
            #{"index_1": "d", "index_2": 1, "column_1": "Bon", "column_2": 99},
            #{"index_1": "d", "index_2": 2, "column_1": "Bam", "column_2": 98},
        ],
        id="range item (list of tuple)"
    ),
    pytest.param(
        slice("b", "c"), 
        [
            {"index_1": "a", "index_2": 1, "column_1": "Jack", "column_2": 10},
            {"index_1": "a", "index_2": 2, "column_1": "John", "column_2": 9},
            #{"index_1": "b", "index_2": 1, "column_1": "James", "column_2": 8},
            #{"index_1": "b", "index_2": 2, "column_1": "Johnny", "column_2": 7},
            #{"index_1": "c", "index_2": 1, "column_1": "Jimmy", "column_2": 8},
            #{"index_1": "c", "index_2": 2, "column_1": "Jim", "column_2": 7},
            {"index_1": "d", "index_2": 1, "column_1": "Bon", "column_2": 99},
            {"index_1": "d", "index_2": 2, "column_1": "Bam", "column_2": 98},
        ],
        id="range item (slice)"
    ),
    pytest.param(
        slice("b", None), 
        [
            {"index_1": "a", "index_2": 1, "column_1": "Jack", "column_2": 10},
            {"index_1": "a", "index_2": 2, "column_1": "John", "column_2": 9},
        ],
        id="range item (slice, left closed)"
    ),
    pytest.param(
        slice(None, "c"), 
        [
            {"index_1": "d", "index_2": 1, "column_1": "Bon", "column_2": 99},
            {"index_1": "d", "index_2": 2, "column_1": "Bam", "column_2": 98},
        ],
        id="range item (slice, right closed)"
    ),
    pytest.param(
        slice(None, None, None), 
        [],
        id="range item (open)"
    ),
    pytest.param(
        ("b", 1), 
        [
            {"index_1": "a", "index_2": 1, "column_1": "Jack", "column_2": 10},
            {"index_1": "a", "index_2": 2, "column_1": "John", "column_2": 9},
            #{"index_1": "b", "index_2": 1, "column_1": "James", "column_2": 8},
            {"index_1": "b", "index_2": 2, "column_1": "Johnny", "column_2": 7},
            {"index_1": "c", "index_2": 1, "column_1": "Jimmy", "column_2": 8},
            {"index_1": "c", "index_2": 2, "column_1": "Jim", "column_2": 7},
            {"index_1": "d", "index_2": 1, "column_1": "Bon", "column_2": 99},
            {"index_1": "d", "index_2": 2, "column_1": "Bam", "column_2": 98},
        ],
        id="single item (tuple)"
    ),
    pytest.param(
        [("b", 1), ("c", 2)], 
        [
            {"index_1": "a", "index_2": 1, "column_1": "Jack", "column_2": 10},
            {"index_1": "a", "index_2": 2, "column_1": "John", "column_2": 9},
            #{"index_1": "b", "index_2": 1, "column_1": "James", "column_2": 8},
            {"index_1": "b", "index_2": 2, "column_1": "Johnny", "column_2": 7},
            {"index_1": "c", "index_2": 1, "column_1": "Jimmy", "column_2": 8},
            #{"index_1": "c", "index_2": 2, "column_1": "Jim", "column_2": 7},
            {"index_1": "d", "index_2": 1, "column_1": "Bon", "column_2": 99},
            {"index_1": "d", "index_2": 2, "column_1": "Bam", "column_2": 98},
        ],
        id="single items (list of tuples)"
    ),
])
def test_delitem_multiindex(table_multiindex, qry, results):
    table = table_multiindex
    del table[qry]
    assert list(table.select()) == results

@pytest.mark.parametrize("qry,results", [
    pytest.param(
        "a", 
        [
            #{"index_1": "a", "column_1": "Jack", "column_2": 10},
            {"index_1": "b", "column_1": "James", "column_2": 8},
            {"index_1": "c", "column_1": "Jimmy", "column_2": 8},
        ],
        id="single item"
    ),
    pytest.param(
        ["a", "b"], 
        [
            #{"index_1": "a", "column_1": "Jack", "column_2": 10},
            #{"index_1": "b", "column_1": "James", "column_2": 8},
            {"index_1": "c", "column_1": "Jimmy", "column_2": 8},
        ],
        id="multi item"
    ),
    pytest.param(
        slice(None, None, None), 
        [],
        id="slice (open)"
    ),
    pytest.param(
        slice("b", None, None),
        [
            {"index_1": "a", "column_1": "Jack", "column_2": 10},
            #{"index_1": "b", "column_1": "James", "column_2": 8},
            #{"index_1": "c", "column_1": "Jimmy", "column_2": 8},
        ],
        id="slice (left closed)"
    ),
    pytest.param(
        slice(None, "b", None),
        [
            #{"index_1": "a", "column_1": "Jack", "column_2": 10},
            #{"index_1": "b", "column_1": "James", "column_2": 8},
            {"index_1": "c", "column_1": "Jimmy", "column_2": 8},
        ],
        id="slice (right closed)"
    ),
])
def test_delitem_singleindex(table_singleindex, qry, results):
    table = table_singleindex
    del table[qry]
    assert list(table.select()) == results

def test_delitem_error(table_multiindex, table_singleindex):

    # Delete single index
    with pytest.raises(KeyError):
        del table_singleindex["not_found"]

    with pytest.raises(KeyError):
        del table_multiindex["not_found"]

    with pytest.raises(KeyError):
        del table_multiindex[("a", -999)]

    with pytest.raises(KeyError):
        del table_multiindex[("zzz",)]

    with pytest.raises(IndexError):
        del table_singleindex[("a", 2)]

    with pytest.raises(IndexError):
        del table_multiindex[("a", 2, "not_found")]


import pytest
from datetime import date

from redbird.sql.expressions import select, Table

def test_enter_commit(engine):
    tbl = Table(bind=engine, name="empty")

    with tbl.transaction() as trans:
        trans.insert(
            {'id': 'a', 'name': 'Johnny', 'birth_date': date(2000, 1, 1), 'score': 100},
        )
        trans.insert(
            {'id': 'b', 'name': 'James', 'birth_date': date(2020, 1, 1), 'score': 100},
        )
    assert [
        {'id': 'a', 'name': 'Johnny', 'birth_date': "2000-01-01", 'score': 100},
        {'id': 'b', 'name': 'James', 'birth_date': "2020-01-01", 'score': 100},
    ] == list(select("select * from empty", bind=engine))

def test_enter_rollback(engine):
    tbl = Table(bind=engine, name="empty")
    try:
        with tbl.transaction() as trans:
            trans.insert(
                {'id': 'a', 'name': 'Johnny', 'birth_date': date(2000, 1, 1), 'score': 100},
            )
            trans.insert(
                {'id': 'b', 'name': 'James', 'birth_date': date(2020, 1, 1), 'score': 100},
            )
            raise RuntimeError("Oops")
    except RuntimeError:
        ...
    assert [] == list(select("select * from empty", bind=engine))

def test_enter_manual_rollback(engine):
    tbl = Table(bind=engine, name="empty")

    with tbl.transaction() as trans:
        trans.insert(
            {'id': 'a', 'name': 'Johnny', 'birth_date': date(2000, 1, 1), 'score': 100},
        )
        trans.insert(
            {'id': 'b', 'name': 'James', 'birth_date': date(2020, 1, 1), 'score': 100},
        )
        trans.rollback()

    assert [] == list(select("select * from empty", bind=engine))

def test_transaction_commit(engine):
    tbl = Table(bind=engine, name="empty")
    trans = tbl.open_transaction()
    trans.insert(
        {'id': 'a', 'name': 'Johnny', 'birth_date': date(2000, 1, 1), 'score': 100},
    )
    trans.insert(
        {'id': 'b', 'name': 'James', 'birth_date': date(2020, 1, 1), 'score': 100},
    )
    trans.commit()

    assert [
        {'id': 'a', 'name': 'Johnny', 'birth_date': "2000-01-01", 'score': 100},
        {'id': 'b', 'name': 'James', 'birth_date': "2020-01-01", 'score': 100},
    ] == list(select("select * from empty", bind=engine))

def test_transaction_rollback(engine):
    tbl = Table(bind=engine, name="empty")
    trans = tbl.open_transaction()
    trans.insert(
        {'id': 'a', 'name': 'Johnny', 'birth_date': date(2000, 1, 1), 'score': 100},
    )
    trans.insert(
        {'id': 'b', 'name': 'James', 'birth_date': date(2020, 1, 1), 'score': 100},
    )
    trans.rollback()

    assert [] == list(select("select * from empty", bind=engine))



import datetime
import sys
import typing
import pytest
from redbird.repos import SQLRepo
from pydantic import BaseModel

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

class MyItem(BaseModel):
    id: str
    name: str
    age: int

class MyItemWithORM(BaseModel):
    id: str
    name: str
    age: int
    class Config:
        orm_mode = True

try:
    from sqlalchemy.orm import declarative_base
    from sqlalchemy import Column, String, Integer
except ImportError:
    pass
else:
    SQLBase = declarative_base()

    class SQLItem(SQLBase):
        __tablename__ = 'pytest'
        id = Column(String, primary_key=True)
        name = Column(String)
        age = Column(Integer)

        def __eq__(self, other):
            if not isinstance(other, SQLItem):
                return False
            return other.id == self.id and other.name == self.name and other.age == self.age

def execute_sql(text, engine):
    import sqlalchemy
    with engine.begin() as conn:
        return conn.execute(sqlalchemy.text(text))

def test_init_engine():
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    engine = create_engine('sqlite://')
    execute_sql(
        """CREATE TABLE pytest (
            id TEXT PRIMARY KEY,
            name TEXT,
            age INTEGER
        )""",
        engine=engine
    )
    repo = SQLRepo(engine=engine, table="pytest")
    assert repo.model_orm.__table__.name == "pytest"
    assert all(hasattr(repo.model_orm, col) for col in ("id", "name", "age"))

    repo.add({"id": "a", "name": "Jack", "age": 500})
    assert list(repo) == [dict(id="a", name="Jack", age=500)]

def test_init_conn_string(tmpdir):
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    import sqlalchemy
    with tmpdir.as_cwd() as old_dir:
        conn_string = 'sqlite:///pytest.db'
        engine = create_engine(conn_string)
        execute_sql(
            """CREATE TABLE pytest (
                id TEXT PRIMARY KEY,
                name TEXT,
                age INTEGER
            )""",
            engine=engine
        )
        repo = SQLRepo(conn_string=conn_string, table="pytest")
        assert repo.model_orm.__table__.name == "pytest"
        assert all(hasattr(repo.model_orm, col) for col in ("id", "name", "age"))

        repo.add({"id": "a", "name": "Jack", "age": 500})
        assert list(repo) == [dict(id="a", name="Jack", age=500)]

def test_init_session():
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    from sqlalchemy.orm import Session
    engine = create_engine('sqlite://')
    session = Session(engine)
    execute_sql("""CREATE TABLE pytest (
        id TEXT PRIMARY KEY,
        name TEXT,
        age INTEGER
    )""", engine=engine)
    repo = SQLRepo(session=session, table="pytest")
    assert repo.model_orm.__table__.name == "pytest"
    assert all(hasattr(repo.model_orm, col) for col in ("id", "name", "age"))

    repo.add({"id": "a", "name": "Jack", "age": 500})
    assert list(repo) == [dict(id="a", name="Jack", age=500)]

def test_init_model():
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    engine = create_engine('sqlite://')
    execute_sql("""CREATE TABLE pytest (
        id TEXT PRIMARY KEY,
        name TEXT,
        age INTEGER
    )""", engine=engine)
    repo = SQLRepo(model=MyItem, engine=engine, table="pytest")
    assert repo.model is MyItem

    repo.add({"id": "a", "name": "Jack", "age": 500})
    assert list(repo) == [MyItem(id="a", name="Jack", age=500)]

def test_init_model_orm_mode():
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    engine = create_engine('sqlite://')
    execute_sql("""CREATE TABLE pytest (
        id TEXT PRIMARY KEY,
        name TEXT,
        age INTEGER
    )""", engine=engine)
    repo = SQLRepo(model=MyItemWithORM, engine=engine, table="pytest")
    assert repo.model is MyItemWithORM

    repo.add({"id": "a", "name": "Jack", "age": 500})
    assert list(repo) == [MyItemWithORM(id="a", name="Jack", age=500)]

def test_init_orm():
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    engine = create_engine('sqlite://')
    execute_sql("""CREATE TABLE pytest (
        id TEXT PRIMARY KEY,
        name TEXT,
        age INTEGER
    )""", engine=engine)
    repo = SQLRepo(model_orm=SQLItem, reflect_model=True, engine=engine)
    assert repo.model_orm is SQLItem
    assert issubclass(repo.model, BaseModel)

    repo.add({"id": "a", "name": "Jack", "age": 500})
    assert list(repo) == [repo.model(id="a", name="Jack", age=500)]

def test_init_model_and_orm():
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    engine = create_engine('sqlite://')
    execute_sql("""CREATE TABLE pytest (
        id TEXT PRIMARY KEY,
        name TEXT,
        age INTEGER
    )""", engine=engine)
    repo = SQLRepo(model=MyItem, model_orm=SQLItem, engine=engine)
    assert repo.model_orm is SQLItem
    assert issubclass(repo.model, BaseModel)

    repo.add({"id": "a", "name": "Jack", "age": 500})
    assert list(repo) == [MyItem(id="a", name="Jack", age=500)]

def test_init_dict():
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    engine = create_engine('sqlite://')
    execute_sql("""CREATE TABLE pytest (
        id TEXT PRIMARY KEY,
        name TEXT,
        age INTEGER
    )""", engine=engine)

    repo = SQLRepo(model=dict, engine=engine, table="pytest")
    repo.add({"id": "a", "name": "Jack", "age": 500})
    assert list(repo) == [{"id": "a", "name": "Jack", "age": 500}]

def test_init_fail_missing_connection():
    pytest.importorskip("sqlalchemy")
    with pytest.raises(TypeError):
        SQLRepo(model=MyItem, table="my_table")

def test_init_deprecated(tmpdir):
    with tmpdir.as_cwd() as old_dir:
        pytest.importorskip("sqlalchemy")
        from sqlalchemy import create_engine
        engine = create_engine('sqlite:///pytest.db')
        execute_sql("""CREATE TABLE pytest (
            id TEXT PRIMARY KEY,
            name TEXT,
            age INTEGER
        )""", engine=engine)

        with pytest.warns(DeprecationWarning):
            repo = SQLRepo.from_engine(model=dict, engine=engine, table="pytest")
        repo.add({"id": "a", "name": "Jack", "age": 500})
        assert list(repo) == [{"id": "a", "name": "Jack", "age": 500}]
        
        with pytest.warns(DeprecationWarning):
            repo = SQLRepo.from_connection_string(model=dict, conn_string="sqlite:///pytest.db", table="pytest")
        repo.add({"id": "b", "name": "Jack", "age": 5000})
        assert list(repo) == [{"id": "a", "name": "Jack", "age": 500}, {"id": "b", "name": "Jack", "age": 5000}]


def test_init_reflect_model_without_primary_key():
    # https://stackoverflow.com/a/23771348/13696660
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    engine = create_engine('sqlite://')
    with pytest.raises(KeyError):
        repo = SQLRepo(model=MyItem, engine=engine, table="mytable", if_missing="create")
    
    execute_sql("""CREATE TABLE pytest (
        id TEXT,
        name TEXT,
        age INTEGER
    )""", engine=engine)
    with pytest.raises(KeyError):
        repo = SQLRepo(model=MyItem, engine=engine, table="mytable")

def test_init_reflect_model_id_field_in_model():
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    engine = create_engine('sqlite://')
    class MyItem(BaseModel):
        __id_field__ = "name"
        name: str
        age: int

    repo = SQLRepo(model=MyItem, engine=engine, table="mytable", if_missing="create")
    assert repo.id_field == "name"

    # Make sure the __id_field__ is not in there
    repo.add(MyItem(name="Jack", age=20))
    obs = execute_sql("select * from mytable", engine=engine)
    assert list(obs) == [('Jack', 20)]

@pytest.mark.parametrize("cls,example_value", [
    pytest.param(str, "Jack", id="str"),
    pytest.param(int, 5, id="int"),
    pytest.param(float, 5.2, id="float"),

    pytest.param(datetime.date, datetime.date(2022, 1, 1), id="datetime.date"),
    pytest.param(datetime.datetime, datetime.datetime(2022, 1, 1, 12, 30), id="datetime.datetime"),
    pytest.param(datetime.timedelta, datetime.timedelta(days=2), id="datetime.timedelta", marks=pytest.mark.skip("SQLite does not support Interval")),

    pytest.param(dict, {"name": "Jack"}, id="dict"),

    pytest.param(typing.Optional[str], "Jack", id="Optional[str]"),
    pytest.param(typing.Union[str, None], "Jack", id="Union[str, None]", marks=pytest.mark.skipif(sys.version_info < (3, 8), reason="<py38")),
    pytest.param(Literal['yes', 'no'], "yes", id="Literal['...', '...']", marks=pytest.mark.skipif(sys.version_info < (3, 8), reason="<py38")),
    pytest.param(Literal[1, 2], 2, id="Literal[1, 2]", marks=pytest.mark.skipif(sys.version_info < (3, 8), reason="<py38")),
])
def test_init_type_insert(cls, example_value):
    pytest.importorskip("sqlalchemy")

    class MyItem(BaseModel):
        __id_field__ = "id"
        id: str
        myfield: cls

    from sqlalchemy import create_engine
    engine = create_engine('sqlite://')
    repo = SQLRepo(model=MyItem, engine=engine, table="mytable", if_missing="create", id_field="id")
    
    repo.add(MyItem(
        id="a",
        myfield=example_value
    ))
    obs = repo.filter_by().all()
    assert obs == [
        MyItem(
            id="a",
            myfield=example_value
        )
    ]

@pytest.mark.parametrize("cls,nullable,sql_type", [
    pytest.param(str, False, 'String', id="str"),
    pytest.param(int, False, 'Integer', id="int"),
    pytest.param(float, False, 'Float', id="float"),

    pytest.param(datetime.date, False, 'Date', id="datetime.date"),
    pytest.param(datetime.datetime, False, 'DateTime', id="datetime.datetime"),
    pytest.param(datetime.timedelta, False, 'Interval', id="datetime.timedelta", marks=pytest.mark.skip("SQLite does not support Interval")),

    pytest.param(dict, False, 'JSON', id="dict"),

    pytest.param(typing.Optional[str], True, 'String', id="Optional[str]"),
    pytest.param(typing.Union[str, None], True, 'String', id="Union[str, None]"),
    pytest.param(typing.Union[None, str], True, 'String', id="Union[None, str]"),
    pytest.param(Literal['yes', 'no'], False, 'String', id="Literal['...', '...']", marks=pytest.mark.skipif(sys.version_info < (3, 8), reason="<py38")),
    pytest.param(Literal[1, 2], False, 'Integer', id="Literal[1, 2]", marks=pytest.mark.skipif(sys.version_info < (3, 8), reason="<py38")),
    pytest.param(typing.Optional[Literal['yes', 'no']], True, 'String', id="Optional[Literal['...', '...']]", marks=pytest.mark.skipif(sys.version_info < (3, 8), reason="<py38")),
])
def test_init_column(cls, nullable, sql_type):
    pytest.importorskip("sqlalchemy")

    class MyItem(BaseModel):
        __id_field__ = "id"
        id: str
        myfield: cls
    import sqlalchemy
    from sqlalchemy import create_engine
    engine = create_engine('sqlite://')
    repo = SQLRepo(model=MyItem, engine=engine, table="mytable", if_missing="create", id_field="id")
    
    # Test column
    column = repo.model_orm.__table__.columns['myfield']
    assert column.nullable is nullable
    assert isinstance(column.type, getattr(sqlalchemy, sql_type))

@pytest.mark.parametrize("cls", [
    pytest.param(Literal[2, 'two'], id="Literal[2, 'two']", marks=pytest.mark.skipif(sys.version_info < (3, 8), reason="<py38")),
    pytest.param(typing.Union[str, int], id="Union[str, int]", marks=pytest.mark.skipif(sys.version_info < (3, 8), reason="<py38")),
    pytest.param(typing.Union[str, int, None], id="Union[str, int, None]", marks=pytest.mark.skipif(sys.version_info < (3, 8), reason="<py38")),
])
def test_init_column_fail(cls):
    pytest.importorskip("sqlalchemy")

    class MyItem(BaseModel):
        __id_field__ = "id"
        id: str
        myfield: cls

    from sqlalchemy import create_engine
    engine = create_engine('sqlite://')

    with pytest.raises(TypeError):
        repo = SQLRepo(model=MyItem, engine=engine, table="mytable", if_missing="create", id_field="id")

def test_init_create_table():
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    from sqlalchemy.exc import NoSuchTableError
    engine = create_engine('sqlite://')
    with pytest.raises(NoSuchTableError):
        repo = SQLRepo(model=MyItem, engine=engine, table="mytable", id_field="id")
    with pytest.raises(NoSuchTableError):
        repo = SQLRepo(model=MyItem, engine=engine, table="mytable", id_field="id", if_missing="raise")

    repo = SQLRepo(model=MyItem, engine=engine, table="mytable", if_missing="create", id_field="id")
    assert issubclass(repo.model, BaseModel)

    repo.add({"id": "a", "name": "Jack", "age": 500})
    assert list(repo) == [MyItem(id="a", name="Jack", age=500)]


import pytest
from datetime import date, datetime

from redbird.sql import insert, select, delete, update, count, execute, Table
from redbird.oper import in_, between


def test_count(engine):
    assert count(table="populated", bind=engine) == 3

@pytest.mark.parametrize("how", [
    "string", "expression"
])
def test_execute(engine, how):
    import sqlalchemy
    if how == "string":
        stmt = """
            INSERT INTO empty(id, name, score, birth_date) 
            VALUES ("a", "Max", 100, "2022-12-31")
        """
    elif how == "expression":
        stmt = sqlalchemy.text("""
            INSERT INTO empty(id, name, score, birth_date) 
            VALUES ("a", "Max", 100, "2022-12-31")
        """)
    execute(stmt, bind=engine)
    
    assert list(select(table="empty", bind=engine)) == [
        {
            "id": "a", "name": "Max", "score": 100, "birth_date": date(2022, 12, 31)
        }
    ]

def test_count_where(engine):
    assert count({"name": "John"}, table="populated", bind=engine) == 1
    assert count({"name": "Johnz"}, table="populated", bind=engine) == 0





import pytest

@pytest.fixture()
def engine(tmpdir):
    sqlalchemy = pytest.importorskip("sqlalchemy")
    engine = sqlalchemy.create_engine(f'sqlite:///{tmpdir}/testing.db')
    with engine.begin() as conn:
        conn.execute(sqlalchemy.text("""
        CREATE TABLE empty (
            id TEXT PRIMARY KEY,
            name TEXT,
            birth_date DATE,
            score INTEGER
        )
        """))

        conn.execute(sqlalchemy.text("""
        CREATE TABLE populated (
            id TEXT PRIMARY KEY,
            name TEXT,
            birth_date DATE,
            score INTEGER
        )
        """))
        conn.execute(sqlalchemy.text("""
        INSERT INTO populated (id, name, birth_date, score)
        VALUES ('a', 'Jack', '2000-01-01', 100),
        ('b', 'John', '1990-01-01', 200),
        ('c', 'James', '2020-01-01', 300)
        """))
    return engine


import argparse

from pydantic import BaseModel, Field
from redbird.repos import SQLRepo

class Item(BaseModel):
    id: int
    name: str = Field(description="Name of the item", default="No name")

repo = SQLRepo(conn_string="sqlite:///example.db", table="items", model=Item, id_field="id", if_missing="create")

def create_item(**kwargs):
    "Create an item"
    item = Item(**kwargs)
    repo.add(item)

def read_item(item_id=None):
    "Read one item (or all)"
    if item_id is None:
        # Read all
        for item in repo:
            print(repr(item))
    else:
        # Read one
        item = repo[item_id]
        print(repr(item))

def update_item(item_id, **values):
    "Update one item"
    values = {key: value for key, value in values.items() if value is not None}
    repo[item_id] = values
    repo.filter_by(id=2).update(**values)

def delete_item(item_id):
    "Delete one item"
    del repo[item_id]

def add_model_to_parser(parser: argparse.ArgumentParser, model):
    "Add Pydantic model's attributes as arguments to the parser"
    fields = model.__fields__
    for name, field in fields.items():
        parser.add_argument(
            f"--{name}", 
            dest=name, 
            type=field.type_, 
            default=field.default,
            help=field.field_info.description,
        )

def parse_args(args=None):
    "Parse CLI arguments"
    parser = argparse.ArgumentParser(prog='Simple CRUD app')
    subparsers = parser.add_subparsers(dest='action')

    # Create subparsers for a single item
    parser_create = subparsers.add_parser("create", help="Create an item")
    parser_read = subparsers.add_parser("read", help="Read an item")
    parser_update = subparsers.add_parser("update", help="Update an item")
    parser_delete = subparsers.add_parser("delete", help="Delete an item")

    # Create subparsers for multiple items
    parser_read_all = subparsers.add_parser("read_all", help="Read items")

    # Add arguments
    parser_read.add_argument("item_id", nargs="?", help="ID of the item")
    for sub_parser in (parser_update, parser_delete):
        sub_parser.add_argument("item_id", help="ID of the item")

    add_model_to_parser(parser_create, model=Item)
    add_model_to_parser(parser_update, model=Item)

    return parser.parse_args(args)

def main(args=None):
    "A simple app to operate on a datastore"
    args = parse_args(args)
    args = vars(args)

    action = args.pop("action")
    func = {
        'create': create_item,
        'read': read_item,
        'update': update_item,
        'delete': delete_item,
    }[action]

    return func(**args)

if __name__ == "__main__":
    main()




from copy import copy
import datetime
from functools import partial
import sys
from typing import TYPE_CHECKING, Callable, Dict, Iterable, List, Mapping, Optional, Tuple, Type, Union
from pathlib import Path
import typing

from redbird.oper import Between, In, Operation, skip
from redbird.packages import sqlalchemy, import_exists

from pydantic import BaseModel


try:
    from typing import Literal
except ImportError: # pragma: no cover
    from typing_extensions import Literal

if TYPE_CHECKING:
    from sqlalchemy.engine import Engine
    from sqlalchemy import Column

SINGULAR = Union[str, int, float, bool, datetime.datetime, datetime.date]

class _KeyInspector:
    # Utility for getitem, delitem and setitem
    def __init__(self, primary_keys, keys):
        self.primary_keys = primary_keys
        self.keys = keys

    def is_multi_item(self):
        return isinstance(self.keys, list)

    def is_specific(self):
        # Is specific item(s) and not ranges with slice or without
        n_primary_keys = len(self.primary_keys)
        if self.is_range():
            return False
        if self.is_multi_item():
            # Keys: [('a', 1), ('a', 2), ...]
            first_key = self.keys[0]
            levels = len(first_key) if not isinstance(first_key, (str,)) else 1
        elif isinstance(self.keys, str):
            levels = 1
        else:
            # Keys: ('a', 1)
            levels = len(self.keys)

        return levels == n_primary_keys

    def is_range(self):
        if isinstance(self.keys, tuple):
            return any(isinstance(key, slice) for key in self.keys)
        else:
            return isinstance(self.keys, slice)

    def _to_equal(self, column, value):
        # Compare column with value (which can be slice)
        if isinstance(value, slice):
            if value.step is not None:
                raise ValueError("Slice step not supported")

            if value.start is None and value.stop is None:
                return sqlalchemy.true()
            elif value.start is None:
                return column <= value.stop
            elif value.stop is None:
                return column >= value.start
            else:
                # Both not None
                return column.between(value.start, value.stop)
        return column == value

    def to_query(self):
        primary_keys = self.primary_keys
        keys = self.keys
        if len(primary_keys) == 0:
            raise TypeError("Table has no primary keys")

        if self.is_multi_item():
            # Fetch multiple items/ranges
            # Fetching multiple items
            qries = []

            n_levels = None
            keys: List[Union[Tuple, SINGULAR]]
            for item in keys:
                if not isinstance(item, tuple):
                    item = (item,)
                n_levels = len(item) if n_levels is None else n_levels
                if len(item) != n_levels:
                    raise IndexError(f"Key out of range")
                item_query = [
                    self._to_equal(primary_keys[i], key)
                    for i, key in enumerate(item)
                ]
                qries.append(sqlalchemy.and_(*item_query))
            qry = sqlalchemy.or_(*qries)
            return qry
        else:
            # Keys: ("a", 1) or "a"
            if not isinstance(keys, tuple):
                keys = (keys,)

            item_queries = []
            for i, key in enumerate(keys):
                item_queries.append(self._to_equal(primary_keys[i], key))
                if isinstance(key, (list, tuple)):
                    raise TypeError(f"Invalid index value type: {type(key)}")
            qry = sqlalchemy.and_(*item_queries)
            return qry


class Table:
    """SQL Table

    Similar to ``sqlalchemy.Table`` except this class always
    points to a specific SQL table in a specific database
    (which may not yet exist) and provides further abstraction.

    Attributes
    ----------
    name : str
        Name of the table.
    bind : sqlalchemy.engine.Connectable
        SQLAlchemy engine or connection.
    
    Examples
    --------
    .. code-block:: python

        from sqlalchemy import create_engine
        from redbird.sql import Table

        table = Table("mytable", bind=create_engine("sqlite://"))
    """
    bind: 'sqlalchemy.engine.Connectable'

    types = {
        str: sqlalchemy.String,
        int: sqlalchemy.Integer,
        float: sqlalchemy.Float,
        bool: sqlalchemy.Boolean,
        datetime.date: sqlalchemy.Date,
        datetime.datetime: sqlalchemy.DateTime,
        datetime.timedelta: sqlalchemy.Interval,
        dict: sqlalchemy.JSON,
    } if import_exists("sqlalchemy") else {}
    _name: str
    _object: 'sqlalchemy.Table'

    class _trans_ctx:
        # Utility for transaction context manager
        def __init__(self, obj:'Table'):
            self.obj = obj

        def __enter__(self):
            new_table = copy(self.obj)
            new_table._ctx = None
            self._ctx = new_table.bind.begin()
            new_table.bind = self._ctx.__enter__()
            return new_table

        def __exit__(self, type_, value, traceback):
            self._ctx.__exit__(type_, value, traceback)

    class _exec_ctx:
        # Utility for execution context manager
        def __init__(self, obj:'Table', args, kwargs):
            self.obj = obj
            self.args = args
            self.kwargs = kwargs

        def __enter__(self):
            bind = self.obj.bind

            # sqlalchemy.engine.Engine does not have "closed" attribute
            # but Connection does. We use this to define if it is open
            is_open = hasattr(bind, "closed") and not bind.closed
            if is_open:
                # "execute" should return Results that are still open
                return self.obj.execute(*self.args, **self.kwargs)
            else:
                self._conn = bind.connect()
                self._conn.__enter__()

                new_table = copy(self.obj)
                new_table.bind = self._conn

                return new_table.execute(*self.args, **self.kwargs)

        def __exit__(self, type_, value, traceback):
            if hasattr(self, "_conn"):
                self._conn.__exit__(type_, value, traceback)

    def __init__(self, name:str, bind:'sqlalchemy.engine.Engine'):
        self.bind = bind
        self.name = name

    def __getitem__(self, keys:Union[SINGULAR, Tuple[SINGULAR], List[Union[SINGULAR, Tuple[SINGULAR]]]]) -> Union[Dict, List[Dict]]:
        "Get item based on the index."
        prim_key_cols = self.object.primary_key.columns
        inspector = _KeyInspector(prim_key_cols, keys=keys)
        qry = inspector.to_query()
        results = list(self.select(qry))

        has_results = len(results) > 0
        is_specific = inspector.is_specific()
        if inspector.is_multi_item():
            if is_specific and not has_results:
                # Not all keys found
                raise KeyError("Missing key(s)")
            return results
        else:
            if not inspector.is_range() and not has_results:
                raise KeyError(f"Item {keys!r} not found")
            if is_specific:
                # Keys point to single item
                return results[0]
            else:
                # Keys point to range
                return results

    def __delitem__(self, keys):
        prim_key_cols = self.object.primary_key.columns
        inspector = _KeyInspector(prim_key_cols, keys)
        qry = inspector.to_query()
        nrows = self.delete(qry)
        if not inspector.is_range() and nrows == 0:
            raise KeyError("Item not found")

    def select(self, qry:Union[str, dict, 'sqlalchemy.sql.ClauseElement', None]=None, columns:Optional[List[str]]=None, parameters:Optional[Dict]=None) -> List[dict]:
        """Read the database table using a query.
        
        Parameters
        ----------
        qry : str, dict, sqlalchemy.sql.ClauseElement, optional
            Query to filter the data. The argument can take various forms:

            - ``str``: Query is considered to be raw SQL
            - ``dict``: Query is considered to be column-filter pairs.
              The pairs are combined using AND operator. If the filter
              is Operation, it is turned to corresponing SQLAlchemy expression.
              Else, the filter is considered to be an equal operator.
            - sqlalchemy expression: The query is considered to be the *where*
              clause of the select query. 

            If not given, all rows are returned.
        columns : list of string, optional
            List of columns to return. By default returns all columns.
        parameters : dict, optional
            Parameters for the query. If the query is as raw SQL,
            they should be set in the query with ``:myparam``, ie.
            ``select * from table where col = :myparam``.

        Returns
        -------
        List of dicts
            Found rows as dicts.

        Examples
        --------
        Select all rows:
        
        .. code-block:: python

            table.select()
        
        Select using raw SQL:
        
        .. code-block:: python

            table.select("SELECT * FROM mytable")

        Select using dictionary:

        .. code-block:: python

            table.select({"column_1": "a value", "column_2": 10})

        .. note::

            The above is same as:

            .. code-block:: sql
            
                SELECT * 
                FROM mytable 
                WHERE column_1 = 'a value' AND column_2 = 10

        Select using SQLAlchemy expressions:

        .. code-block:: python

            from sqlalchemy import Column
            table.select((Column("column_1") == "a value") & (Column("column_2") == 10))

        .. note::

            The above is same as:

            .. code-block:: sql
            
                SELECT * 
                FROM mytable 
                WHERE column_1 = 'a value' AND column_2 = 10

        Select and return specific column(s):

        .. code-block:: python

            table.select(columns=["column_1", "column_2"])

        .. note::

            The above is same as:

            .. code-block:: sql
            
                SELECT column_1, column_2 
                FROM mytable

        Select using raw strings and SQL parameters:

        .. code-block:: python

            table.select(
                "SELECT * FROM mytable WHERE column_1 = :myparam",
                parameters={"myparam": "a value"}
            )
        """
        if isinstance(qry, Path):
            qry = qry.read_text()
        elif qry is None:
            qry = sqlalchemy.true()

        if isinstance(qry, dict):
            qry = to_expression(qry)
        if isinstance(qry, str):
            statement = sqlalchemy.text(qry)
        elif isinstance(qry, (sqlalchemy.sql.Selectable, sqlalchemy.sql.elements.TextClause)):
            statement = qry
        else:
            if columns is None:
                columns = ()
            else:
                columns = [
                    self.object.columns[col]
                    for col in columns
                ]
            where = qry
            statement = self.object.select(*columns)
            statement = statement.where(where)
            statement = statement.select_from(self.object)

        if parameters is not None:
            statement = statement.bindparams(**parameters)

        with self.execution(statement) as results:
            rows = results.mappings()
            if self.name is not None:
                rows = self._format_results(rows)
            return list(rows)
    
    def insert(self, data:Union[Mapping, List[dict]]):
        """Insert data to the database.
        
        Parameters
        ----------
        data : dict, list of dicts
            Data to be inserted.

        Examples
        --------
        Insert a single row:
        
        .. code-block:: python

            table.insert({"column_1": "a", "column_2": 1})

        Insert multiple rows:
        
        .. code-block:: python

            table.insert([
                {"column_1": "a", "column_2": 1},
                {"column_1": "b", "column_2": 2},
            ])
        """
        table = self.object
        if isinstance(data, Mapping):
            statement = table.insert().values(**data)
            self.execute(statement)
        else:
            self.execute(sqlalchemy.insert(self.object), data)

    def delete(self, where:Union[dict, 'sqlalchemy.sql.ClauseElement']) -> int:
        """Delete row(s) from the table.
        
        Parameters
        ----------
        where : dict, sqlalchemy expression
            Where clause to delete data.

        Returns
        -------
        int
            Count of rows deleted.

        Examples
        --------

        Delete using dictionary:

        .. code-block:: python

            table.delete({"column_1": "a", "column_2": 1})

        .. note::

            The above is same as:

            .. code-block:: sql
            
                DELETE FROM mytable 
                WHERE column_1 = 'a' AND column_2 = 1

        Delete using SQL expressions:

        .. code-block:: python

            from sqlalchemy import Column
            table.delete((Column("column_1") == "a") & (Column("column_2") == 1))

        .. note::

            The above is same as:

            .. code-block:: sql
            
                DELETE FROM mytable 
                WHERE column_1 = 'a' AND column_2 = 1

        Delete all:

        .. code-block:: python

            table.delete({})

        """
        if isinstance(where, dict):
            where = to_expression(where)
        table = self.object
        statement = table.delete().where(where)
        result = self.execute(statement)
        return result.rowcount

    def update(self, where:Union[dict, 'sqlalchemy.sql.ClauseElement'], values:dict) -> int:
        """Update row(s) in the table.
        
        Parameters
        ----------
        where : dict, sqlalchemy expression
            Where clause to update rows.
        values : dict
            Column-value pairs to update the 
            rows matching the where clause.

        Returns
        -------
        int
            Count of rows updated.

        Examples
        --------

        Update using dicts:

        .. code-block:: python

            table.update({"column_1": "a", "column_2": 1}, {"column_3": "new value"})

        .. note::

            The above is same as:

            .. code-block:: sql
                
                UPDATE mytable 
                SET column_3='new value' 
                WHERE column_1 = 'a' and column_2 = 1

        Update using expressions:

        .. code-block:: python

            from sqlalchemy import Column
            table.update((Column("column_1") == "a") & (Column("column_2") == 1), {"column_3": "new value"})

        .. note::

            The above is same as:

            .. code-block:: sql
                
                UPDATE mytable 
                SET column_3='new value' 
                WHERE column_1 = 'a' and column_2 = 1

        Update all:

        .. code-block:: python

            table.update({}, {"column_3": "new value"})

        """
        if isinstance(where, dict):
            where = to_expression(where)
        table = self.object
        statement = table.update().where(where).values(values)
        result = self.execute(statement)
        return result.rowcount

    def count(self, where=None) -> int:
        """Count the number of rows.
        
        Parameters
        ----------
        where : dict, sqlalchemy expression, optional
            Where clause to be satisfied for counting the rows.

        Returns
        -------
        int
            Count of rows (satisfying the where clause).

        Examples
        --------

        Count based on dict:

        .. code-block:: python

            table.count({"column_1": "a", "column_2": 1})
        """
        stmt = sqlalchemy.select(sqlalchemy.func.count()).select_from(self.object)
        if where is not None:
            if isinstance(where, dict):
                where = to_expression(where)
            stmt = stmt.where(where)

        with self.execution(stmt) as results:
            return results.scalar_one()

    def _format_results(self, res:Iterable[Tuple]) -> Iterable[dict]:
        columns = self._get_types()
        for row in res:
            row = {name: conv(row[name]) for name, conv in columns.items()}
            yield row

    def _get_types(self) -> Dict[str, Callable]:
        "Get table's column types"
        cols = dict(self.object.columns)
        return {col_name: partial(to_native, sql_type=col.type, nullable=col.nullable) for col_name, col in cols.items()}

    def _to_sqlalchemy_type(self, cls):
        from sqlalchemy.sql import sqltypes
        if isinstance(cls, sqltypes.TypeEngine):
            return cls
        is_older_py = sys.version_info < (3, 8)
        origin = typing.get_origin(cls) if not is_older_py else None
        if origin is not None:
            # In form: 
            # - Literal['', '']
            # - Optional[...]
            args = typing.get_args(cls)
            if origin is typing.Union:
                # Either:
                # - Union[...]
                # - Optional[...]
                # Only Union[<TYPE>, NoneType] is allowed
                none_type = type(None)
                has_none_type = none_type in args
                if len(args) > 2 or (len(args) == 2 and not has_none_type):
                    raise TypeError(f"Union has more than one optional type: {str(cls)}. Cannot define SQL data type")
                # Get the non-None type
                for arg in args:
                    if arg is not none_type:
                        cls = arg
                        break
            
            if origin is Literal:
                type_ = type(args[0])
                for arg in args[1:]:
                    if not isinstance(arg, type_):
                        raise TypeError(f"Literal values are not same types: {str(cls)}. Cannot define SQL data type")
                cls = type_
        return self.types.get(cls)

    def reflect(self):
        """Reflect the table from the database."""
        self._object = reflect_table(self.name, bind=self.bind)

    def drop(self):
        """Drop the table."""
        self.object.drop(bind=self.bind)

    def exists(self) -> bool:
        """Check if the table exists."""
        inspector = sqlalchemy.inspect(self.bind)
        return inspector.has_table(self.name)

    def create(self, columns:Union[List['Column'], Mapping[str, Type]], exist_ok=False):
        """Create the table
        
        Parameters
        ----------
        columns : dict, list of sqlalchemy.Column, dict or string
            Columns to be created.
        exist_ok : bool
            If false (default), an exception is raised.

        Examples
        --------

        There are various ways to call this method:

        - Using list of columns (all of the)

        Create a table with columns ``column_1``, ``column_2`` and ``column_3``
        (all of them are textual columns): 

        .. code-block:: python

            table.create(["column_1", "column_2", "column_3"])

        Create a table with columns ``column_1``, ``column_2`` and ``column_3``
        with varying data types: 

        .. code-block:: python

            import datetime
            table.create({"column_1": str, "column_2": int, "column_3": datetime.datetime})

        Create a table with columns ``column_1``, ``column_2`` and ``column_3``
        using SQLAlchemy columns: 

        .. code-block:: python

            from sqlalchemy import Column, String, Integer, DateTime
            table.create([
                Column("column_1", type_=String()),
                Column("column_2", type_=Integer()),
                Column("column_3", type_=DateTime())
            ])

        Create a table with columns ``column_1``, ``column_2`` and ``column_3``
        using list of dicts: 

        .. code-block:: python

            from sqlalchemy import DateTime
            table.create([
                {"name": "column_1", "type_": str},
                {"name": "column_2", "type_": int},
                {"name": "column_3", "type_": DateTime()},
            ])
        """
        if isinstance(columns, Mapping):
            columns = [
                sqlalchemy.Column(name, self._to_sqlalchemy_type(type_))
                for name, type_ in columns.items()
            ]
        elif isinstance(columns, (list, tuple)):
            columns = [
                sqlalchemy.Column(col, type_=sqlalchemy.String()) if isinstance(col, str) 
                else sqlalchemy.Column(
                    type_=self._to_sqlalchemy_type(col['type_']), 
                    **{k: v for k, v in col.items() if k not in ('type_',)}
                ) if isinstance(col, dict)
                else col
                for col in columns
            ]
        meta = sqlalchemy.MetaData()
        tbl = sqlalchemy.Table(self.name, meta, *columns)

        if exist_ok and self.exists():
            self._object = tbl
            return

        tbl.create(bind=self.bind)
        self._object = tbl

    def create_from_model(self, model:BaseModel, primary_column=None):
        if isinstance(primary_column, str):
            primary_column = (primary_column,)
        sql_cols = [
            sqlalchemy.Column(
                name, 
                self._to_sqlalchemy_type(field.type_), 
                primary_key=name in primary_column if primary_column is not None else False, 
                nullable=not field.required,
                default=field.default
            )
            for name, field in model.__fields__.items()
        ]
        self.create(sql_cols)
  
    def execute(self, *args, **kwargs):
        """Execute SQL statement or raw SQL.
        
        Parameters
        ----------
        *args : tuple
            Passed directly to sqlalchemy.Connection.execute.
        **args : dict
            Passed directly to sqlalchemy.Connection.execute.
        """
        if len(args) == 1 and isinstance(args[0], str):
            # SQLAlchemy v2.0 won't accept string as 
            # expression but we do
            args = (sqlalchemy.text(*args),)
        conn = self.bind
        if isinstance(conn, sqlalchemy.engine.Engine):
            with conn.begin() as conn:
                return conn.execute(*args, **kwargs)
        else:
            return conn.execute(*args, **kwargs)

    def open_transaction(self):
        """Open a transaction.
        
        Examples
        --------
        .. code-block:: python

            from redbird.sql import Table
            from sqlalchemy import create_engine

            table = Table("mytable", bind=create_engine(...))

            # Open the transaction
            transaction = table.open_transaction()

            # Perform operations
            transaction.insert({"col_1": "a", "col_2": "b"})
            transaction.delete({"col_2": "c"})

            # Commit or rollback the changes
            if successful:
                transaction.commit()
            else:
                transaction.rollback()
        """
        self._ctx = self._trans_ctx(self)
        return self._ctx.__enter__()

    def transaction(self):
        """Open a transaction context.
        
        If an error is raised inside the with block,
        the changes are rollbacked. Else they are 
        commited.

        Examples
        --------
        .. code-block:: python

            from redbird.sql import Table
            from sqlalchemy import create_engine

            table = Table("mytable", bind=create_engine(...))

            with table.transaction() as trans:

                # Perform operations
                trans.insert({"col_1": "a", "col_2": "b"})
                trans.delete({"col_2": "c"})
        """
        return self._trans_ctx(self)

    def rollback(self):
        "Rollback the open transaction (a transaction must be open)"
        return self.bind.get_transaction().rollback()

    def commit(self):
        "Commit the open transaction (a transaction must be open)"
        return self.bind.get_transaction().commit()

    def execution(self, *args, **kwargs):
        "Context manager to manipulate the execution result"
        return self._exec_ctx(self, args, kwargs)

    @property
    def object(self) -> 'sqlalchemy.Table':
        "sqlalchemy.Table: SQLAlchemy representation of the table"
        if self._object is None:
            self.reflect()
        return self._object

    @object.setter
    def object(self, value: 'sqlalchemy.Table'):
        self._object = value
        self._name = value.name

    @property
    def name(self) -> str:
        "str: name of the table"
        if self._object is not None:
            return self._object.name
        return self._name
    
    @name.setter
    def name(self, value: str):
        self._name = value
        self._object = None


def to_native(value, sql_type, nullable=False):
    "Convert sql type to Python native."
    py_type = sql_type.python_type
    if isinstance(value, py_type):
        return value
    elif py_type is datetime.datetime and isinstance(value, str):
        return datetime.datetime.fromisoformat(value)
    elif py_type is datetime.date and isinstance(value, str):
        return datetime.date.fromisoformat(value)
    if nullable and value is None:
        return value
    return py_type(value)

def create_table(*args, bind:'Engine', table:str, **kwargs):
    """Create a table to a SQL database.
    
    Parameters
    ----------
    bind : sqlalchemy.engine.Engine
        SQLAlchemy engine for the connection
    table : str
        Name of the table to be created.
    *args
        Passed to :py:meth:`redbird.sql.Table.create`
    **kwargs
        Passed to :py:meth:`redbird.sql.Table.create`
    """
    return Table(bind=bind, name=table).create(*args, **kwargs)

def reflect_table(table:str, *columns, bind:'Engine', meta=None):
    """Reflect a table in an SQL database"""
    if meta is None:
        meta = sqlalchemy.MetaData()
    return sqlalchemy.Table(table, meta, *columns, autoload_with=bind)

def select(*args, bind:'Engine', table:str=None, **kwargs):
    """Read rows from a table in a SQL database.
    
    Parameters
    ----------
    engine : sqlalchemy.engine.Engine
        SQLAlchemy engine for the connection
    table : str
        Name of the table to use.
    *args
        Passed to :py:meth:`redbird.sql.Table.select`
    **kwargs
        Passed to :py:meth:`redbird.sql.Table.select`
    """
    return Table(bind=bind, name=table).select(*args, **kwargs)

def insert(*args, bind:'Engine', table:str=None, **kwargs):
    """Insert row(s) to a table in a SQL database.

    Parameters
    ----------
    bind : sqlalchemy.engine.Engine
        SQLAlchemy engine for the connection
    table : str
        Name of the table to use.
    *args
        Passed to :py:meth:`redbird.sql.Table.insert`
    **kwargs
        Passed to :py:meth:`redbird.sql.Table.insert`
    """
    return Table(bind=bind, name=table).insert(*args, **kwargs)

def delete(*args, bind:'Engine', table:str=None, **kwargs):
    """Delete row(s) in a table in a SQL database.
    
    Parameters
    ----------
    bind : sqlalchemy.engine.Engine
        SQLAlchemy engine for the connection
    table : str
        Name of the table to use.
    *args
        Passed to :py:meth:`redbird.sql.Table.delete`
    **kwargs
        Passed to :py:meth:`redbird.sql.Table.delete`
    """
    return Table(bind=bind, name=table).delete(*args, **kwargs)

def update(*args, bind:'Engine', table:str=None, **kwargs):
    """Update row(s) to a table in a SQL database.
    
    Parameters
    ----------
    bind : sqlalchemy.engine.Engine
        SQLAlchemy engine for the connection
    table : str
        Name of the table to use.
    *args
        Passed to :py:meth:`redbird.sql.Table.update`
    **kwargs
        Passed to :py:meth:`redbird.sql.Table.update`
    """
    return Table(bind=bind, name=table).update(*args, **kwargs)

def count(*args, bind:'Engine', table:str=None, **kwargs):
    """Count the number of rows in a table in a SQL database.

    Parameters
    ----------
    bind : sqlalchemy.engine.Engine
        SQLAlchemy engine for the connection
    table : str
        Name of the table to use.
    *args
        Passed to :py:meth:`redbird.sql.Table.count`
    **kwargs
        Passed to :py:meth:`redbird.sql.Table.count`
    """
    return Table(bind=bind, name=table).count(*args, **kwargs)

def execute(*args, bind:'Engine', **kwargs):
    """Execute raw SQL or a SQL expression in a SQL database.

    Parameters
    ----------
    bind : sqlalchemy.engine.Engine
        SQLAlchemy engine for the connection
    table : str
        Name of the table to use.
    *args
        Passed to :py:meth:`redbird.sql.Table.execute`
    **kwargs
        Passed to :py:meth:`redbird.sql.Table.execute`
    """
    return Table(bind=bind, name=None).execute(*args, **kwargs)


def to_expression(qry:dict, table=None):
    stmt = sqlalchemy.true()
    for column_name, oper_or_value in qry.items():
        column = getattr(table, column_name) if table is not None else sqlalchemy.Column(column_name)
        if isinstance(oper_or_value, Operation):
            oper = oper_or_value
            if isinstance(oper, Between):
                sql_oper = column.between(oper.start, oper.end)
            elif isinstance(oper, In):
                sql_oper = column.in_(oper.value)
            elif oper is skip:
                continue
            elif hasattr(oper, "__py_magic__"):
                magic = oper.__py_magic__
                oper_method = getattr(column, magic)

                # Here we form the SQLAlchemy operation, ie.: column("mycol") >= 5
                sql_oper = oper_method(oper.value)
            else:
                raise NotImplementedError(f"Not implemented operator: {oper}")
        elif isinstance(oper_or_value, slice):
            start, stop, step = oper_or_value.start, oper_or_value.stop, oper_or_value.step
            if step is not None:
                raise ValueError("Step cannot be interpret")
            if start is not None and stop is not None:
                sql_oper = column.between(start, stop)
            elif start is not None:
                sql_oper = column >= start
            elif stop is not None:
                sql_oper = column <= stop
            else:
                sql_oper = sqlalchemy.true()
        else:
            value = oper_or_value
            sql_oper = column == value
        stmt &= sql_oper
    return stmt


import sys

sys.stderr.write("""
Unsupported installation method: python setup.py
Please use `python -m pip install .` instead.
"""
)
#sys.exit(1)
from setuptools import setup

setup(
    name="redbird",
    install_requires = ["pydantic", "typing_extensions"],
)




from .base import BaseRepo, BaseResult

try:
    from ._version import *
except ImportError:
    # Package was not built the standard way
    __version__ = version = '0.0.0.UNKNOWN'
    __version_tuple__ = version_tuple = (0, 0, 0, 'UNKNOWN', '')

def create_repo(model=None, model_orm=None, repo_name="MemoryRepo"):
    registry[repo_name](model=model, model_orm=model_orm)



from typing import TYPE_CHECKING
from redbird.utils.importing import import_optional, import_exists

if TYPE_CHECKING:
    import sqlalchemy as sqlalchemy_lib
    import pymongo as pymongo_lib
    import requests as requests_lib
    import pydantic_sqlalchemy as pydantic_sqlalchemy_lib

sqlalchemy: 'sqlalchemy_lib' = import_optional("sqlalchemy")
pymongo: 'pymongo_lib' = import_optional("pymongo")
requests: 'requests_lib' = import_optional("requests")
pydantic_sqlalchemy: 'pydantic_sqlalchemy_lib' = import_optional("pydantic_sqlalchemy")